@@ -182,6 +182,7 @@ struct _MetaDisplay
   Atom atom_net_moveresize_window;
   Atom atom_net_desktop_geometry;
   Atom atom_net_desktop_viewport;
+  Atom atom_metacity_version;
 
   /* This is the actual window from focus events,
    * not the one we last set
@@ -265,6 +266,7 @@ struct _MetaDisplay
   MetaResizePopup *grab_resize_popup;
   GTimeVal    grab_last_moveresize_time;
   Time        grab_motion_notify_time;
+  GList*      grab_old_window_stacking;
   
   /* we use property updates as sentinels for certain window focus events
    * to avoid some race conditions on EnterNotify events
