@@ -2,6 +2,7 @@
 
 /* 
  * Copyright (C) 2001 Havoc Pennington
+ * Copyright (C) 2003 Rob Adams
  * 
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -26,9 +27,20 @@
 #include <X11/Xatom.h>
 #include <string.h>
 
-void meta_workspace_queue_calc_showing  (MetaWorkspace *workspace);
+void meta_workspace_queue_calc_showing   (MetaWorkspace *workspace);
+static void set_active_space_hint        (MetaScreen *screen);
+static void focus_ancestor_or_mru_window (MetaWorkspace *workspace,
+                                          MetaWindow    *not_this_one,
+                                          Time           timestamp);
 
-static void set_active_space_hint      (MetaScreen *screen);
+static void
+maybe_add_to_list (MetaScreen *screen, MetaWindow *window, gpointer data)
+{
+  GList **mru_list = data;
+
+  if (window->on_all_workspaces)
+    *mru_list = g_list_prepend (*mru_list, window);
+}
 
 MetaWorkspace*
 meta_workspace_new (MetaScreen *screen)
@@ -42,6 +54,7 @@ meta_workspace_new (MetaScreen *screen)
     g_list_append (workspace->screen->workspaces, workspace);
   workspace->windows = NULL;
   workspace->mru_list = NULL;
+  meta_screen_foreach_window (screen, maybe_add_to_list, &workspace->mru_list);
 
   workspace->work_areas = NULL;
   workspace->work_areas_invalid = TRUE;
@@ -54,6 +67,8 @@ meta_workspace_new (MetaScreen *screen)
   workspace->right_struts = NULL;
   workspace->top_struts = NULL;
   workspace->bottom_struts = NULL;
+
+  workspace->showing_desktop = FALSE;
   
   return workspace;
 }
@@ -79,7 +94,7 @@ meta_workspace_free (MetaWorkspace *workspace)
 
       /* pop front of list we're iterating over */
       meta_workspace_remove_window (workspace, window);
-      g_assert (window->workspaces != NULL);
+      g_assert (window->workspace != NULL);
 
       tmp = next;
     }
@@ -110,21 +125,21 @@ void
 meta_workspace_add_window (MetaWorkspace *workspace,
                            MetaWindow    *window)
 {
-  g_return_if_fail (!meta_workspace_contains_window (workspace, window));
+  g_return_if_fail (window->workspace == NULL);
   
   /* If the window is on all workspaces, we want to add it to all mru
    * lists, otherwise just add it to this workspaces mru list
    */
   if (window->on_all_workspaces) 
     {
-      if (window->workspaces == NULL)
+      if (window->workspace == NULL)
         {
           GList* tmp = window->screen->workspaces;
           while (tmp)
             {
               MetaWorkspace* work = (MetaWorkspace*) tmp->data;
               if (!g_list_find (work->mru_list, window))
-                work->mru_list = g_list_append (work->mru_list, window);
+                work->mru_list = g_list_prepend (work->mru_list, window);
 
               tmp = tmp->next;
             }
@@ -133,11 +148,11 @@ meta_workspace_add_window (MetaWorkspace *workspace,
   else
     {
       g_assert (g_list_find (workspace->mru_list, window) == NULL);
-      workspace->mru_list = g_list_append (workspace->mru_list, window);
+      workspace->mru_list = g_list_prepend (workspace->mru_list, window);
     }
 
   workspace->windows = g_list_prepend (workspace->windows, window);
-  window->workspaces = g_list_prepend (window->workspaces, workspace);
+  window->workspace = workspace;
 
   meta_window_set_current_workspace_hint (window);
   
@@ -160,10 +175,10 @@ void
 meta_workspace_remove_window (MetaWorkspace *workspace,
                               MetaWindow    *window)
 {
-  g_return_if_fail (meta_workspace_contains_window (workspace, window));
+  g_return_if_fail (window->workspace == workspace);
 
   workspace->windows = g_list_remove (workspace->windows, window);
-  window->workspaces = g_list_remove (window->workspaces, workspace);
+  window->workspace = NULL;
 
   /* If the window is on all workspaces, we don't want to remove it
    * from the MRU list unless this causes it to be removed from all 
@@ -171,16 +186,13 @@ meta_workspace_remove_window (MetaWorkspace *workspace,
    */
   if (window->on_all_workspaces) 
     {
-      if (window->workspaces == NULL)
+      GList* tmp = window->screen->workspaces;
+      while (tmp)
         {
-          GList* tmp = window->screen->workspaces;
-          while (tmp)
-            {
-              MetaWorkspace* work = (MetaWorkspace*) tmp->data;
-              work->mru_list = g_list_remove (work->mru_list, window);
+          MetaWorkspace* work = (MetaWorkspace*) tmp->data;
+          work->mru_list = g_list_remove (work->mru_list, window);
 
-              tmp = tmp->next;
-            }
+          tmp = tmp->next;
         }
     }
   else
@@ -235,13 +247,6 @@ meta_workspace_relocate_windows (MetaWorkspace *workspace,
   g_assert (workspace->windows == NULL);
 }
 
-gboolean
-meta_workspace_contains_window (MetaWorkspace *workspace,
-                                MetaWindow    *window)
-{
-  return g_list_find (workspace->windows, window) != NULL;
-}
-
 void
 meta_workspace_queue_calc_showing  (MetaWorkspace *workspace)
 {
@@ -258,9 +263,11 @@ meta_workspace_queue_calc_showing  (MetaWorkspace *workspace)
 
 void
 meta_workspace_activate_with_focus (MetaWorkspace *workspace,
-                                    MetaWindow    *focus_this)
+                                    MetaWindow    *focus_this,
+                                    Time           timestamp)
 {
   MetaWorkspace *old;
+  MetaWindow *move_window;
   
   meta_verbose ("Activating workspace %d\n",
                 meta_workspace_index (workspace));
@@ -268,36 +275,73 @@ meta_workspace_activate_with_focus (MetaWorkspace *workspace,
   if (workspace->screen->active_workspace == workspace)
     return;
 
+  /* Note that old can be NULL; e.g. when starting up */
   old = workspace->screen->active_workspace;
   
   workspace->screen->active_workspace = workspace;
 
   set_active_space_hint (workspace->screen);
 
+  /* If the "show desktop" mode is active for either the old workspace
+   * or the new one *but not both*, then update the
+   * _net_showing_desktop hint
+   */
+  if (old && (old->showing_desktop ^ workspace->showing_desktop))
+    meta_screen_update_showing_desktop_hint (workspace->screen);
+
   if (old == NULL)
     return;
 
+  move_window = NULL;
+  if (workspace->screen->display->grab_op == META_GRAB_OP_MOVING ||
+      workspace->screen->display->grab_op == META_GRAB_OP_KEYBOARD_MOVING)
+    move_window = workspace->screen->display->grab_window;
+      
+  if (move_window != NULL)
+    {
+      if (move_window->on_all_workspaces)
+        move_window = NULL; /* don't move it after all */
+
+      /* We put the window on the new workspace, flip spaces,
+       * then remove from old workspace, so the window
+       * never gets unmapped and we maintain the button grab
+       * on it.
+       */
+      if (move_window && (move_window->workspace != workspace))
+        {
+          meta_workspace_remove_window (workspace, move_window);
+          meta_workspace_add_window (workspace, move_window);
+        }
+    }
+
   meta_workspace_queue_calc_showing (old);
   meta_workspace_queue_calc_showing (workspace);
 
+  if (move_window)
+      /* Removes window from other spaces */
+      meta_window_change_workspace (move_window, workspace);
+
   if (focus_this)
     {
-      meta_window_focus (focus_this,
-                         meta_display_get_current_time (focus_this->display));
+      meta_window_focus (focus_this, timestamp);
       meta_window_raise (focus_this);
     }
+  else if (move_window)
+    {
+      meta_window_raise (move_window);
+    }
   else
     {
       meta_topic (META_DEBUG_FOCUS, "Focusing default window on new workspace\n");
-      meta_workspace_focus_default_window (workspace, NULL);
+      meta_workspace_focus_default_window (workspace, NULL, timestamp);
     }
 }
 
 void
-meta_workspace_activate (MetaWorkspace *workspace)
+meta_workspace_activate (MetaWorkspace *workspace,
+                         Time           timestamp)
 {
-  meta_workspace_activate_with_focus (workspace,
-                                      NULL);
+  meta_workspace_activate_with_focus (workspace, NULL, timestamp);
 }
 
 int
@@ -340,7 +384,7 @@ meta_workspace_list_windows (MetaWorkspace *workspace)
     {
       MetaWindow *window = tmp->data;
 
-      if (meta_window_visible_on_workspace (window, workspace))
+      if (meta_window_located_on_workspace (window, workspace))
         workspace_windows = g_list_prepend (workspace_windows,
                                             window);
 
@@ -746,18 +790,83 @@ meta_workspace_get_name (MetaWorkspace *workspace)
 
 void
 meta_workspace_focus_default_window (MetaWorkspace *workspace,
-                                     MetaWindow *not_this_one)
+                                     MetaWindow    *not_this_one,
+                                     Time           timestamp)
 {
+  if (timestamp == CurrentTime)
+    {
+      meta_warning ("CurrentTime used to choose focus window; "
+                    "focus window may not be correct.\n");
+    }
+
+
   if (meta_prefs_get_focus_mode () == META_FOCUS_MODE_CLICK)
-    meta_workspace_focus_mru_window (workspace, not_this_one);
+    focus_ancestor_or_mru_window (workspace, not_this_one, timestamp);
   else
-    meta_screen_focus_mouse_window (workspace->screen, not_this_one);
+    {
+      MetaWindow * window;
+      window = meta_screen_get_mouse_window (workspace->screen, not_this_one);
+      if (window &&
+          window->type != META_WINDOW_DOCK &&
+          window->type != META_WINDOW_DESKTOP)
+        {
+          if (timestamp == CurrentTime)
+            {
+
+              /* We would like for this to never happen.  However, if
+               * it does happen then we kludge since using CurrentTime
+               * can mean ugly race conditions--and we can avoid these
+               * by allowing EnterNotify events (which come with
+               * timestamps) to handle focus.
+               */
+
+              meta_topic (META_DEBUG_FOCUS,
+                          "Not focusing mouse window %s because EnterNotify events should handle that\n", window->desc);
+            }
+          else
+            {
+              meta_topic (META_DEBUG_FOCUS,
+                          "Focusing mouse window %s\n", window->desc);
+              meta_window_focus (window, timestamp);
+            }
+
+          if (workspace->screen->display->autoraise_window != window &&
+              meta_prefs_get_auto_raise ()) 
+            {
+              meta_display_queue_autoraise_callback (workspace->screen->display,
+                                                     window);
+            }
+        }
+      else if (meta_prefs_get_focus_mode () == META_FOCUS_MODE_SLOPPY)
+        focus_ancestor_or_mru_window (workspace, not_this_one, timestamp);
+      else if (meta_prefs_get_focus_mode () == META_FOCUS_MODE_MOUSE)
+        {
+          meta_topic (META_DEBUG_FOCUS,
+                      "Setting focus to no_focus_window, since no valid "
+                      "window to focus found.\n");
+          meta_display_focus_the_no_focus_window (workspace->screen->display,
+                                                  timestamp);
+        }
+    }
 }
 
-/* Focus MRU window (or top window if failed) on active workspace */
-void
-meta_workspace_focus_mru_window (MetaWorkspace *workspace,
-                                 MetaWindow *not_this_one)
+static gboolean
+record_ancestor (MetaWindow *window,
+                 void       *data)
+{
+  MetaWindow **result = data;
+
+  *result = window;
+  return FALSE; /* quit with the first ancestor we find */
+}
+
+/* Focus ancestor of not_this_one if there is one, otherwise focus the MRU
+ * window on active workspace
+ */
+static void
+focus_ancestor_or_mru_window (MetaWorkspace *workspace,
+                              MetaWindow    *not_this_one,
+                              Time           timestamp)
 {
   MetaWindow *window = NULL;
   GList *tmp;
@@ -765,12 +874,43 @@ meta_workspace_focus_mru_window (MetaWorkspace *workspace,
   if (not_this_one)
     meta_topic (META_DEBUG_FOCUS,
                 "Focusing MRU window excluding %s\n", not_this_one->desc);
-  
+  else
+    meta_topic (META_DEBUG_FOCUS,
+                "Focusing MRU window\n");
+
+  /* First, check to see if we need to focus an ancestor of a window */  
+  if (not_this_one)
+    {
+      MetaWindow *ancestor;
+      ancestor = NULL;
+      meta_window_foreach_ancestor (not_this_one, record_ancestor, &ancestor);
+      if (ancestor != NULL)
+        {
+          meta_topic (META_DEBUG_FOCUS,
+                      "Focusing %s, ancestor of %s\n", 
+                      ancestor->desc, not_this_one->desc);
+      
+          meta_window_focus (ancestor, timestamp);
+
+          /* Also raise the window if in click-to-focus */
+          if (meta_prefs_get_focus_mode () == META_FOCUS_MODE_CLICK)
+            meta_window_raise (ancestor);
+
+          return;
+        }
+    }
+
+  /* No ancestor, look for the MRU window */
   tmp = workspace->mru_list;  
 
   while (tmp)
     {
-      if (((MetaWindow*) tmp->data) != not_this_one)
+      MetaWindow* tmp_window;
+      tmp_window = ((MetaWindow*) tmp->data);
+      if (tmp_window != not_this_one           &&
+          !tmp_window->minimized               &&
+          tmp_window->type != META_WINDOW_DOCK &&
+          tmp_window->type != META_WINDOW_DESKTOP)
         {
           window = tmp->data;
 	  break;
@@ -784,8 +924,7 @@ meta_workspace_focus_mru_window (MetaWorkspace *workspace,
       meta_topic (META_DEBUG_FOCUS,
                   "Focusing workspace MRU window %s\n", window->desc);
       
-      meta_window_focus (window,
-                         meta_display_get_current_time (workspace->screen->display));
+      meta_window_focus (window, timestamp);
 
       /* Also raise the window if in click-to-focus */
       if (meta_prefs_get_focus_mode () == META_FOCUS_MODE_CLICK)
@@ -793,41 +932,8 @@ meta_workspace_focus_mru_window (MetaWorkspace *workspace,
     }
   else
     {
-      meta_topic (META_DEBUG_FOCUS, "No MRU window to focus found\n");
-      meta_workspace_focus_top_window (workspace, not_this_one);
-    }
-}
-
-/* Focus top window on workspace */
-void
-meta_workspace_focus_top_window (MetaWorkspace *workspace,
-                                 MetaWindow    *not_this_one)
-{
-  MetaWindow *window;
-
-  if (not_this_one)
-    meta_topic (META_DEBUG_FOCUS,
-                "Focusing top window excluding %s\n", not_this_one->desc);
-  
-  window = meta_stack_get_default_focus_window (workspace->screen->stack,
-                                                workspace,
-                                                not_this_one);
-
-  /* FIXME I'm a loser on the CurrentTime front */
-  if (window)
-    {
-      meta_topic (META_DEBUG_FOCUS,
-                  "Focusing top window %s\n", window->desc);
-
-      meta_window_focus (window, 
-	                 meta_display_get_current_time (workspace->screen->display));
-
-      /* Also raise the window if in click-to-focus */
-      if (meta_prefs_get_focus_mode () == META_FOCUS_MODE_CLICK)
-        meta_window_raise (window);
-    }
-  else
-    {
-      meta_topic (META_DEBUG_FOCUS, "No top window to focus found\n");
+      meta_topic (META_DEBUG_FOCUS, "No MRU window to focus found; focusing no_focus_window.\n");
+      meta_display_focus_the_no_focus_window (workspace->screen->display,
+                                              timestamp);
     }
 }
